/****************************************************************************

    This file is part of The Newtonator soft synth.

    Copyright (C) 2010-2011 Michael Bechard <mbechard@users.sourceforge.net>

    The Newtonator is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The Newtonator is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with The Newtonator.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************/

#include "newt_parent_widget.h"
#include "config.h"

using namespace sigc;
using namespace Gtk;

NewtParentWidget::NewtParentWidget()
: m_divAdjust1(0, -3, 3)
, m_divAdjust2(0, -3, 3)

, m_tab1a(4, 1)
, m_tab1b(2, 4)
, m_tab1c(1, 2)
, m_tab1d(1, 4)
, m_tab1e(2, 3)
, m_frame1a("Amp. Env.")
, m_frame1b("Misc.")
, m_frame1c("Clip Wobble")
, m_frame1d("Velocity")
, m_mainEnvAtk("Attack", .001, 10, .01, this, c_amp_attack)
, m_mainEnvDec("Decay", .001, 10, .1, this, c_amp_decay)
, m_mainEnvSus("Sustain", .001, 1, .8, this, c_amp_sustain)
, m_mainEnvRel("Release", .001, 10, .3, this, c_amp_release)
, m_velocReset("Velocity Reset", -1.1, 1, 1, this, c_veloc_reset)
, m_velocDelay("Velocity Delay", 0, 100, 0, this, c_veloc_delay, 0, 1)
, m_ceilingBounce("Ceiling Bounce", -1, 1, 0, this, c_ceil_bounce)
, m_wobbleA("A", 0, 1, .499, this, c_wobble_a)
, m_wobbleB("B", 0, .1, .02, this, c_wobble_b, 4, .001)
, m_wobbleScale("Scale", 0, .7, .05, this, c_wobble_scale)
, m_wobbleTimbre("Timbre", 0, 1, .5, this, c_wobble_timbre)
, m_velocRM("Velocity Ring Mod.", 0, 1, 0, this, c_veloc_rm)
, m_velocSaH("Velocity Sample And Hold", 0, 1000, 0, this, c_veloc_samphold, 0, 1)
, m_amp("Overall Amplitude", 0, 1, .8, this, c_amp)
, m_ceilingBehavior("Ceiling Behavior", this, c_ceiling_behav)
, m_stereo("Stereo")

, m_tab2a(3, 1)
, m_tab2b(2, 4)
, m_tab2c(2, 3)
, m_tab2d(2, 4)
, m_frame2a("Envelope")
, m_frame2b("Misc.")
, m_frame2c("Frequency")
, m_gmodEnvAtk("Attack", .001, 10, .01, this, c_gmod_attack)
, m_gmodEnvDec("Decay", .001, 10, .1, this, c_gmod_decay)
, m_gmodEnvSus("Sustain", .001, 1, .8, this, c_gmod_sustain)
, m_gmodEnvRel("Release", .001, 10, .3, this, c_gmod_release)
, m_gravScale("Gravity Scaling", 0, 3, 0, this, c_grav_scale)
, m_gmodDepth("Depth", 0, 10, 0, this, c_gmod_depth)
, m_gmodDCOffset("DC Offset", .01, 10, 1, this, c_gmod_dcoffset)
, m_gmodFreqAbs("Absolute Frequency", 0, 20000, 100, this, c_gmod_freq)
, m_gmodFreqDrift("Freq. Drift", -.5, .5, 0, this, c_gmod_freq_drift)
, m_gmodDeadZone("Dead Zone", 0, 1, 0, this, c_dead_zone)
, m_gmodEnv("Use Envelope")
, m_gmodFollowFreq("Follow Frequency")
, m_gmodWave("Grav. Mod. Waveform", this, c_gmod_wav)
, m_gmodFreqDiv(m_divAdjust1)
, m_gmodFreqDivLab("Freq. Divider")
{
    const guint spacing1 = 15;

    InitModulatorFreqs();
    m_hideValueChange = false;
    signal_value_should_change().connect(sigc::mem_fun(*this, &NewtParentWidget::on_value_should_change));

    //Page 1
    append_page(m_tab1a, "Main", "Main");
    m_tab1a.set_spacings(spacing1);
    //amp. envelope
    m_tab1a.attach(m_frame1a, 0, 1, 0, 1);
    m_frame1a.add(m_tab1b);
    m_tab1b.attach(m_amp, 0, 2, 0, 1);
    m_tab1b.attach(m_mainEnvAtk, 0, 1, 1, 2);
    m_tab1b.attach(m_mainEnvDec, 1, 2, 1, 2);
    m_tab1b.attach(m_mainEnvSus, 2, 3, 1, 2);
    m_tab1b.attach(m_mainEnvRel, 3, 4, 1, 2);
    //clip wobble
    m_tab1a.attach(m_frame1c, 0, 1, 1, 2);
    m_frame1c.add(m_tab1d);
    m_tab1d.attach(m_wobbleScale, 0, 1, 0, 1);
    m_tab1d.attach(m_wobbleA, 1, 2, 0, 1);
    m_tab1d.attach(m_wobbleB, 2, 3, 0, 1);
    m_tab1d.attach(m_wobbleTimbre, 3, 4, 0, 1);
    //velocity
    m_tab1a.attach(m_frame1d, 0, 1, 2, 3);
    m_frame1d.add(m_tab1e);
    m_tab1e.attach(m_velocReset, 1, 2, 0, 1);
    m_tab1e.attach(m_velocDelay, 0, 1, 1, 2);
    m_tab1e.attach(m_velocSaH, 1, 2, 1, 2);
    m_tab1e.attach(m_velocRM, 2, 3, 1, 2);
    //misc
    m_tab1a.attach(m_frame1b, 0, 1, 3, 4);
    m_frame1b.add(m_tab1c);
    m_tab1c.attach(m_ceilingBehavior, 0, 1, 0, 1);
    m_tab1c.attach(m_ceilingBounce, 1, 2, 0, 1);
    m_tab1c.attach(m_stereo, 0, 1, 1, 2);
    InitCBComboBox();

    //Page 1 - control sensitivity
    m_wobbleSensitivity.Add(&m_wobbleA);
    m_wobbleSensitivity.Add(&m_wobbleB);
    m_wobbleSensitivity.Add(&m_wobbleTimbre);

    //Page 1 - event wiring
    m_ceilingBehavior.signal_changed().connect(sigc::mem_fun(*this, &NewtParentWidget::on_ceilingBehavior_changed));
    m_stereo.signal_toggled().connect(sigc::mem_fun(*this, &NewtParentWidget::on_stereo_toggled));
    m_wobbleScale.signal_value_changed().connect(sigc::mem_fun(*this, &NewtParentWidget::on_wobbleScale_value_changed));

    //Page 2 - layout
    append_page(m_tab2a, "Gravity Mod.", "Gravity Mod.");
    m_tab2a.set_spacings(spacing1);
    m_tab2a.attach(m_frame2a, 0, 1, 0, 1);
    m_frame2a.add(m_tab2b);
    m_tab2b.attach(m_gmodEnv, 0, 1, 0, 1);
    m_tab2b.attach(m_gmodEnvAtk, 0, 1, 1, 2);
    m_tab2b.attach(m_gmodEnvDec, 1, 2, 1, 2);
    m_tab2b.attach(m_gmodEnvSus, 2, 3, 1, 2);
    m_tab2b.attach(m_gmodEnvRel, 3, 4, 1, 2);
    m_tab2a.attach(m_frame2c, 0, 1, 1, 2);
    m_frame2c.add(m_tab2d);
    m_tab2d.attach(m_gmodFollowFreq, 0, 1, 0, 1);
    m_tab2d.attach(m_gmodFreqDrift, 0, 1, 1, 2);
    m_tab2d.attach(m_gmodFreqDivLab, 1, 2, 1, 2);
    m_tab2d.attach(m_gmodFreqDiv, 2, 3, 1, 2);
    m_tab2d.attach(m_gmodFreqAbs, 3, 4, 1, 2);
    m_tab2a.attach(m_frame2b, 0, 1, 2, 3);
    m_frame2b.add(m_tab2c);
    m_tab2c.attach(m_gravScale, 0, 1, 0, 1);
    m_tab2c.attach(m_gmodDepth, 1, 2, 0, 1);
    m_tab2c.attach(m_gmodDCOffset, 2, 3, 0, 1);
    m_tab2c.attach(m_gmodWave, 0, 1, 1, 2);
    m_tab2c.attach(m_gmodDeadZone, 1, 2, 1, 2);
    InitWFComboBox(m_gmodWave);

    //Page 2 - control sensitivity
    m_gmodEnvSensitivity.Add(&m_gmodEnvAtk);
    m_gmodEnvSensitivity.Add(&m_gmodEnvDec);
    m_gmodEnvSensitivity.Add(&m_gmodEnvSus);
    m_gmodEnvSensitivity.Add(&m_gmodEnvRel);
    m_gmodEnvFollowYesSensitivity.Add(&m_gmodFreqDiv);
    m_gmodEnvFollowYesSensitivity.Add(&m_gmodFreqDrift);
    m_gmodEnvFollowNoSensitivity.Add(&m_gmodFreqAbs);
    m_gmodDepthSensitivity.Add(&m_gmodEnvSensitivity);
    m_gmodDepthSensitivity.Add(&m_gmodEnvFollowYesSensitivity);
    m_gmodDepthSensitivity.Add(&m_gmodEnvFollowNoSensitivity);
    m_gmodDepthSensitivity.Add(&m_gmodEnv);
    m_gmodDepthSensitivity.Add(&m_gmodFollowFreq);
    m_gmodDepthSensitivity.Add(&m_gmodWave);
    m_gmodEnvSensitivity.SetSensitivity(false);
    SetGModFreqFollow(false);
    m_gmodDepthSensitivity.SetSensitivity(false);

    //Page 2 - event wiring
    m_gmodEnv.signal_toggled().connect(sigc::mem_fun(*this, &NewtParentWidget::on_gmodEnv_toggled));
    m_gmodFollowFreq.signal_toggled().connect(sigc::mem_fun(*this, &NewtParentWidget::on_gmodFollowFreq_toggled));
    m_gmodFreqDiv.signal_value_changed().connect(sigc::mem_fun(*this, &NewtParentWidget::on_gmodFreqDiv_value_changed));
    m_gmodDepth.signal_value_changed().connect(sigc::mem_fun(*this, &NewtParentWidget::on_gmodDepth_value_changed));
}

NewtParentWidget::~NewtParentWidget()
{
    //dtor
    TRACE("NewtParentWidget::~NewtParentWidget()");
}

void NewtParentWidget::ChangeValue(ControlEnum ctrl, const stk::StkFloat val)
{
    TRACE2("NewtParentWidget::ChangeValue", ctrl, val);

    m_hideValueChange = true;
    signal_value_should_change().emit(ctrl, val);
    m_hideValueChange = false;
}

void NewtParentWidget::on_value_should_change(ControlEnum ctrl, stk::StkFloat val)
{
    switch (ctrl)
    {
        //check boxes
        case c_stereo:
            m_stereo.set_active((val == 1.0));
            break;
        case c_gmod_use_env:
            m_gmodEnv.set_active((val == 1.0));
            break;
        case c_gmod_follow_freq:
            m_gmodFollowFreq.set_active((val == 1.0));
            break;

        //HScales
        case c_gmod_freq_div:
            m_gmodFreqDiv.set_value(val);
            break;

        default:
            //don't care, subclassed controls will subscribe to this event
            break;
    }
}

void NewtParentWidget::EmitValueChange(ControlEnum ctrlID, stk::StkFloat val)
{
    if (!m_hideValueChange)
        m_valueHasChanged.emit(ctrlID, val);
}

void NewtParentWidget::SetGModFreqFollow(bool val)
{
    m_gmodEnvFollowNoSensitivity.SetSensitivity(!val);
    m_gmodEnvFollowYesSensitivity.SetSensitivity(val);
}

void NewtParentWidget::InitCBComboBox()
{
    CeilingBehavEnum ids[] = { ExtClip, IntClip, Bounce };
    const char* names[] = { "Ext. Clip", "Int. Clip", "Bounce" };

    m_ceilingBehavior.AddItems(ids, names, 3);
    m_ceilingBehavior.SetActive(ExtClip);
    m_ceilingBounce.set_sensitive(false);
}

void NewtParentWidget::InitWFComboBox(SelectionWidget<VariSource::Type>& comboBox)
{
    #ifdef INCLUDE_EXT_INPUTS
    VariSource::Type ids[] = { VariSource::SINE, VariSource::SAW, VariSource::SQUARE, VariSource::IMPULSE, VariSource::NOISE, VariSource::EXTERNAL };
    const char* names[] = { "Sine", "Saw", "Square", "Impulse", "Noise", "External" };
    const int count = 6;
    #else
    VariSource::Type ids[] = { VariSource::SINE, VariSource::SAW, VariSource::SQUARE, VariSource::IMPULSE, VariSource::NOISE };
    const char* names[] = { "Sine", "Saw", "Square", "Impulse", "Noise" };
    const int count = 5;
    #endif

    comboBox.AddItems(ids, names, count);
    comboBox.SetActive(VariSource::SINE);
}

void NewtParentWidget::InitModulatorFreqs()
{
    m_gmodFreqDiv.set_digits(0);
}

// BEGIN EVENTS //

void NewtParentWidget::on_gmodDepth_value_changed()
{
    if (m_gmodDepth.GetValue() == 0.0)
        m_gmodDepthSensitivity.SetSensitivity(false);
    else
        m_gmodDepthSensitivity.SetSensitivity(true);
}

void NewtParentWidget::on_wobbleScale_value_changed()
{
    if (m_wobbleScale.GetValue() == 0)
        m_wobbleSensitivity.SetSensitivity(false);
    else
        m_wobbleSensitivity.SetSensitivity(true);
}

void NewtParentWidget::on_ceilingBehavior_changed()
{
    CeilingBehavEnum enumVal = m_ceilingBehavior.GetActive();

    m_ceilingBounce.set_sensitive((enumVal == Bounce || enumVal == BounceMirror));
}

void NewtParentWidget::on_gmodEnv_toggled()
{
    m_gmodEnvSensitivity.SetSensitivity(m_gmodEnv.get_active());
    EmitValueChange(c_gmod_use_env, (m_gmodEnv.get_active() ? 1.0 : 0.0));
}


void NewtParentWidget::on_gmodFollowFreq_toggled()
{
    SetGModFreqFollow(m_gmodFollowFreq.get_active());
    EmitValueChange(c_gmod_follow_freq, (m_gmodFollowFreq.get_active() ? 1.0 : 0.0));
}

void NewtParentWidget::on_stereo_toggled()
{
    EmitValueChange(c_stereo, (m_stereo.get_active() ? 1.0 : 0.0));
}

void NewtParentWidget::on_gmodFreqDiv_value_changed()
{
    EmitValueChange(c_gmod_freq_div, m_gmodFreqDiv.get_value());
}
// END EVENTS //
